#pragma once


#pragma intrinsic(memcpy, memset)

void* zalloc(size_t size);
void zfree(void* ptr);
void* aligned_zalloc(size_t size);
void aligned_zfree(void* memblock);
void* aligned_zrealloc(void* memblock, size_t size);





class imodel : public IDSIMMODEL {
public:
    virtual BYTE model_id() = 0;
};



/********************************************************************/


class I8275 : public imodel, public IMSGHLR {
public:
    int isdigital(CHAR* pinname);
    void setup(IINSTANCE* inst, IDSIMCKT* dsim);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA* data);
    void simulate(ABSTIME time, DSIMMODES mode);
    void callback(ABSTIME time, EVENTID eventid);
    LRESULT msghlr(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam);
    DWORD Thread();
    LRESULT WndProg(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam);
    BYTE model_id();
    ~I8275();
private:
    void pin_CS_onchange(ABSTIME time, DSIMMODES mode);
    void pin_WR_onchange(ABSTIME time, DSIMMODES mode);
    void pin_RD_onchange(ABSTIME time, DSIMMODES mode);
    void pin_CLK_onchange(ABSTIME time, DSIMMODES mode);
    BYTE get_data();
    void set_LC04(ABSTIME time, BYTE tmp);
    void set_CC06(ABSTIME time, BYTE tmp);
    void set_data(ABSTIME time, BYTE tmp);
    void set_data_Z(ABSTIME time);
    void draw();

    ULONG size_x = 512;
    ULONG size_y = 384;

    void* pbuff = nullptr;
    BITMAPINFO* bmi = nullptr;
    IUSERPOPUP* pup = nullptr;

    HANDLE ev = 0;
    HANDLE th = 0;
    HWND hw = 0;
    HWND hwt = 0;


    

    char redraw = 0;
    BYTE iclk = 0;
    BYTE logg = 0;
    BYTE cursor_blink_cnt = 0; //    

    BYTE screen_comp1_reg = 0; //  
    BYTE screen_comp2_reg = 0; //  
    BYTE screen_comp3_reg = 0; //  
    BYTE screen_comp4_reg = 0; //  

    BYTE takt = 0; //  ,       
    BYTE SREG_reg = 0; //   
    BYTE lp_char_numb_reg = 0; //   ( )
    BYTE lp_row_numb_reg = 0;  //   ( )

    BYTE cur_char_numb_reg = 0; //  
    BYTE cur_row_numb_reg = 0;  //  
    BYTE display_cmd = 0; //       
    BYTE hrtc_cnt = 0; //    

    BYTE LC04_cnt = 0; //     
    BYTE row_cnt = 0;  //     ( )  
    BYTE char_cnt = 0; //     ( )  
    BYTE current_row = 0; // 0/1 -       ,   

    BYTE dma_burst_cnt = 0;  //     DMA    
    BYTE clk_between_dma_cnt = 0; //      DMA 
    BYTE reserved8 = 0;
    BYTE reserved81 = 0;
    
    BYTE row_buf_cnt[2]; // /    
    BYTE row_fifo_cnt[2]; // /    
    BYTE row_fifo[16][2]; //0..15 atributs       ( )
    BYTE row_buf[80][2]; //0..79 chars       


    PAINTSTRUCT ps;


    IINSTANCE* instance;
    //    IDSIMCKT* ckt;

    IDSIMPIN* pin_D[8];
    IDSIMPIN* pin_CLK;
    IDSIMPIN* pin_cs;
    IDSIMPIN* pin_nRD;
    IDSIMPIN* pin_nWR;
    IDSIMPIN* pin_A0;
    IDSIMPIN* pin_DRQ;
    IDSIMPIN* pin_nDACK;
    IDSIMPIN* pin_IRQ;
    IDSIMPIN* pin_LPEN;
    IDSIMPIN* pin_LTEN;
    IDSIMPIN* pin_LC0;
    IDSIMPIN* pin_LC1;
    IDSIMPIN* pin_LC2;
    IDSIMPIN* pin_LC3;
    IDSIMPIN* pin_LC4;
    IDSIMPIN* pin_CC0;
    IDSIMPIN* pin_CC1;
    IDSIMPIN* pin_CC2;
    IDSIMPIN* pin_CC3;
    IDSIMPIN* pin_CC4;
    IDSIMPIN* pin_CC5;
    IDSIMPIN* pin_CC6;
    IDSIMPIN* pin_LA0;
    IDSIMPIN* pin_LA1;
    IDSIMPIN* pin_VSP;
    IDSIMPIN* pin_HLGT;
    IDSIMPIN* pin_RVV;
    IDSIMPIN* pin_GPA0;
    IDSIMPIN* pin_GPA1;
    IDSIMPIN* pin_HRTC;
    IDSIMPIN* pin_VRTC;


};


/********************************************************************/



class RAM : public imodel {
public:
    int isdigital(CHAR* pinname);
    void setup(IINSTANCE* inst, IDSIMCKT* dsim);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA* data);
    void simulate(ABSTIME time, DSIMMODES mode);
    void callback(ABSTIME time, EVENTID eventid);
    BYTE model_id();
    BYTE nv = 0;
    ~RAM();
private:
    void pin_CE_onchange(ABSTIME time, DSIMMODES mode);
    void pin_OE_onchange(ABSTIME time, DSIMMODES mode);
    void pin_WE_onchange(ABSTIME time, DSIMMODES mode);
    BYTE get_data();
    unsigned get_addr();
    void set_data(ABSTIME time, BYTE tmp);
    void set_data_Z(ABSTIME time);
    BYTE* pram = nullptr;
    IINSTANCE* instance;
    HANDLE file = INVALID_HANDLE_VALUE;
    //    IDSIMCKT* ckt;
    IDSIMPIN* pin_ce;
    IDSIMPIN* pin_oe;
    IDSIMPIN* pin_we;
    IDSIMPIN* pin_d[8];
    IDSIMPIN* pin_a[11];
};


/********************************************************************/



class I8255 : public imodel {
public:
    void setup(IINSTANCE* inst, IDSIMCKT* dsim);
    int isdigital(CHAR* pinname);
    void runctrl(RUNMODES mode);
    void actuate(REALTIME time, ACTIVESTATE newstate);
    BOOL indicate(REALTIME time, ACTIVEDATA* data);
    void simulate(ABSTIME time, DSIMMODES mode);
    void callback(ABSTIME time, EVENTID eventid);
    BYTE model_id();
private:
    void pin_CS_onchange(ABSTIME time, DSIMMODES mode);
    void pin_WR_onchange(ABSTIME time, DSIMMODES mode);
    void pin_RD_onchange(ABSTIME time, DSIMMODES mode);
    void pin_RESET_onchange(ABSTIME time, DSIMMODES mode);
    void set_data_Z(ABSTIME time, unsigned port);
    void set_data_Z_pc(ABSTIME time, BYTE mask);
    BYTE get_data(unsigned port);
    void set_data(ABSTIME time, BYTE data, unsigned port);
    void set_data_pc(ABSTIME time, BYTE data, BYTE mask);

    BYTE pa;
    BYTE pb;
    BYTE pc;
    BYTE cw;
    LONG logging;

    IINSTANCE* instance;
    //IDSIMCKT* ckt;

    IDSIMPIN* pin_cs;
    IDSIMPIN* pin_rd;
    IDSIMPIN* pin_wr;
    IDSIMPIN* pin_reset;
    IDSIMPIN* pin_a0;
    IDSIMPIN* pin_a1;
    IDSIMPIN* pins[32];
};

