// dllmain.cpp : Defines the entry point for the DLL application.

#ifndef _DEBUG
#pragma comment(linker, "/ENTRY:DllMain")
#endif


#define WIN32_LEAN_AND_MEAN             /* Exclude rarely-used stuff from Windows headers */

#include <intrin.h>
#include <windows.h>
#include <shlwapi.h>

#pragma comment(lib, "shlwapi.lib")


#include "vsm.hpp"
#include "Header.h"


#pragma optimize("s", on)


HANDLE heap;

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved) {
    if (ul_reason_for_call == DLL_PROCESS_ATTACH) {
        DisableThreadLibraryCalls(hModule);
        heap = HeapCreate(0, 0, 0);
        if (!heap) return false;
    }
    else if (ul_reason_for_call == DLL_PROCESS_DETACH) {
        HeapDestroy(heap);
    }
    return true;
}


void* zalloc(size_t size) {
    return HeapAlloc(heap, HEAP_ZERO_MEMORY, size);
}

void zfree(void* ptr) {
    if (ptr) {
        if (HeapValidate(heap, 0, ptr)) {
            HeapFree(heap, 0, ptr);
            return;
        }
        MessageBoxA(0, "Heap corrupt or Invalid ptr.", "ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    }
}


// must be a power of two
#define MEMALIGN 32

#define ALIGN_PTR(ptr) ((void*)((((DWORD_PTR)((char*)ptr + MEMALIGN + sizeof(void*)))& ~(MEMALIGN - 1))))
#define SAVED_PTR(x) ((void**)((DWORD_PTR)((char*)x - sizeof(void*)) & ~(sizeof(void*) - 1)))

void* aligned_zalloc(size_t size) {
    void* ptmp = zalloc(size + MEMALIGN + sizeof(void*));
    if (!ptmp) return 0;
    void* memblock = ALIGN_PTR(ptmp);
    void** saved = SAVED_PTR(memblock);
    *saved = ptmp;
    return memblock;
}


void aligned_zfree(void* memblock) {
    if (memblock) {
        void** saved = SAVED_PTR(memblock);
        zfree(*saved);
    }
}

void* aligned_zrealloc(void* memblock, size_t size) {
    if (memblock) {
        void** saved = SAVED_PTR(memblock);
        void* ptmp = HeapReAlloc(heap, HEAP_REALLOC_IN_PLACE_ONLY, *saved, size + MEMALIGN + sizeof(void*));
        if (ptmp) {
            __stosd((unsigned long*)memblock, 0, size / 4);
            return memblock;
        }
        zfree(*saved);
    }
    return aligned_zalloc(size);
}

void* __cdecl operator new (size_t size) {
    return HeapAlloc(heap, 0, size);
}

void __cdecl operator delete(void* ptr, size_t size) {
    if (HeapSize(heap, 0, ptr) != size) {
        MessageBoxA(0, "Wrong memory size specified when calling operator delete.", "ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    }
    else {
        zfree(ptr);
    }
}

void __cdecl operator delete(void* ptr) {
    zfree(ptr);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const DWORD d[5] = { 2155905153, 2147483648, 2155905152, 2290649224, 0 };
extern "C" IDSIMMODEL __declspec(dllexport) * __cdecl createdsimmodel(CHAR * device, ILICENCESERVER * ils) {
    for (unsigned i = 0; i < 5; i++) if (ils->authorize(d[i])) break;
    if (StrStrIA(device, "55")) return new I8255;
    if (StrStrIA(device, "75")) return new I8275;
    if (StrStrIA(device, "video")) return new RAM;
    if (StrStrIA(device, "nvram")) {
        RAM* r = new RAM;
        if (r) r->nv = 1;
        return r;
    }
    char str[MAX_PATH + 32];
    wvnsprintfA(str, MAX_PATH + 32, "Unknown model \"%s\" to load.", va_list(&device));
    MessageBoxA(GetForegroundWindow(), str, "ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    return nullptr;
}


extern "C" VOID __declspec(dllexport) __cdecl deletedsimmodel(IDSIMMODEL * model) {
    BYTE r = ((imodel*)model)->model_id();
    if (r == 16) delete (RAM*)model;
    else if (r == 55) delete (I8255*)model;
    else if (r == 75) delete (I8275*)model;
    else MessageBoxA(GetForegroundWindow(), "Unknown model to delete.", "ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
}


extern "C" int __cdecl _purecall() {
    DWORD_PTR r = (DWORD_PTR)_ReturnAddress();
    RaiseException(STATUS_ACCESS_VIOLATION, EXCEPTION_NONCONTINUABLE, 1, &r);
    return 0;
}

class type_info {
    virtual ~type_info();
};

type_info::~type_info() {
    _purecall();
}