#define WIN32_LEAN_AND_MEAN             /* Exclude rarely-used stuff from Windows headers */

#include <intrin.h>
#include <windows.h>   
#include "vsm.hpp"   
#include "Header.h"



#define num_of_ports_pins 32
#define DB 0
#define PA 8
#define PB 16
#define PC 24
#define PLO 0x0F
#define PHI 0xF0
#define PA_IN 0b00010000
#define PB_IN 0b00000010
#define PCL_IN 0b00000001
#define PCH_IN 0b00001000




/********************************************************************/


#pragma optimize("s", on)

BYTE I8255::model_id() {
    return 55;
}
     

/********************************************************************/

void I8255::setup(IINSTANCE* inst, IDSIMCKT* dsim) {
    instance = inst;
    //ckt = dsim;
    pa = 0; 
    pb = 0; 
    pc = 0; 
    cw = 0x9B;
    logging = inst->getinitval("LOGGING", 8);

    pin_cs = instance->getdsimpin("$CS$", true); 
    pin_rd = instance->getdsimpin("$RD$", true); 
    pin_wr = instance->getdsimpin("$WR$", true); 
    pin_reset = instance->getdsimpin("RESET", true);   
    pin_a0 = instance->getdsimpin("A0", true);   
    pin_a1 = instance->getdsimpin("A1", true);   
    
    char pname[4] = "D0\0";
    for (unsigned i = DB; i < DB + 8; i++, pname[1]++) pins[i] = instance->getdsimpin(pname, true);
    memcpy(pname, "PA0", 4);
    for (unsigned i = PA; i < PA + 8; i++, pname[2]++) pins[i] = instance->getdsimpin(pname, true);
    memcpy(pname, "PB0", 4);
    for (unsigned i = PB; i < PB + 8; i++, pname[2]++) pins[i] = instance->getdsimpin(pname, true);
    memcpy(pname, "PC0", 4);
    for (unsigned i = PC; i < PC + 8; i++, pname[2]++) pins[i] = instance->getdsimpin(pname, true);

    pin_cs->sethandler(this, (PINHANDLERFN)&I8255::pin_CS_onchange);
    pin_rd->sethandler(this, (PINHANDLERFN)&I8255::pin_RD_onchange);
    pin_wr->sethandler(this, (PINHANDLERFN)&I8255::pin_WR_onchange);
    pin_reset->sethandler(this, (PINHANDLERFN)&I8255::pin_RESET_onchange);
}


int I8255::isdigital(CHAR* pinname) {
    return 1;
}

BOOL I8255::indicate(REALTIME time, ACTIVEDATA *data) {      
    return false;
}   

#pragma optimize("t", on)   

void I8255::runctrl(RUNMODES mode)  {   
}   
       
void I8255::actuate(REALTIME time, ACTIVESTATE newstate) {   

}   

void I8255::set_data_Z(ABSTIME time, unsigned port) {
    for (unsigned i = port; i < port + 8; i++) pins[i]->drivetristate(time);
}

void I8255::set_data_Z_pc(ABSTIME time, BYTE mask) {
    BYTE b = 1;
    for (unsigned i = PC; i < PC + 8; i++, b *= 2) if (b & mask) pins[i]->drivetristate(time);
}

BYTE I8255::get_data(unsigned port) {
    BYTE tmp = 0;
    BYTE b = 1;
    for (unsigned i = port; i < port + 8; i++, b *= 2) if (!islow(pins[i]->istate())) tmp |= b;
    return tmp;
}

void I8255::set_data(ABSTIME time, BYTE data, unsigned port) {
    BYTE b = 1;
    for (unsigned i = port; i < port + 8; i++, b *= 2) pins[i]->drivebool(time, b & data);
}

void I8255::set_data_pc(ABSTIME time, BYTE data, BYTE mask) {
    BYTE b = 1;
    for (unsigned i = PC; i < PC + 8; i++, b *= 2) if (b & mask) pins[i]->drivebool(time, b & data);
}

void I8255::pin_RESET_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_reset->isposedge()) {
        for (unsigned i = 0; i < num_of_ports_pins; i++) pins[i]->drivetristate(time);
        pa = 0;
        pb = 0;
        pc = 0;
        cw = 0x9B;
        if (logging) instance->log("i8255: reset active.", pa);
    }
    else if (pin_reset->isnegedge()) {
        if (logging) instance->log("i8255: reset inactive.", pa);
    }
}

void I8255::pin_CS_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_reset->isinactive()) return;
    if (pin_cs->isposedge()) set_data_Z(time, DB);
}

void I8255::pin_WR_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_reset->isinactive()) return;
    if (!pin_cs->isinactive()) return;
    if (pin_wr->isposedge()) {
        BYTE addr = 0;
        if (pin_a0->isactive()) addr |= 1;
        if (pin_a1->isactive()) addr |= 2;
        BYTE datain = get_data(DB);
        if (addr == 0) {
            pa = datain;
            if ((cw & PA_IN) == 0) set_data(time, pa, PA);
            if (logging & 1) instance->log("i8255: write %s: 0x%02X", "PA", pa);
        }
        else if (addr == 1) {
            pb = datain;
            if ((cw & PB_IN) == 0) set_data(time, pb, PB);
            if (logging & 2) instance->log("i8255: write %s: 0x%02X", "PB", pb);
        }
        else if (addr == 2) {
            pc = datain;
            if ((cw & PCL_IN) == 0) set_data_pc(time, pc, PLO);
            if ((cw & PCH_IN) == 0) set_data_pc(time, pc, PHI);
            if (logging & 4) instance->log("i8255: write %s: 0x%02X", "PC", pc);
        }
        else if (datain & 0x80) {
            cw = datain;
            if (cw & 0b01100100) {
                instance->warning("i8255: unsupported mode, control word: 0x%02X", cw);
                return;
            }
            if (logging) instance->log("i8255: set control word: 0x%02X", cw);
            if (cw & PA_IN) set_data_Z(time, PA); else set_data(time, pa, PA); 
            if (cw & PB_IN) set_data_Z(time, PB); else set_data(time, pb, PB);
            if (cw & PCL_IN) set_data_Z_pc(time, PLO); else set_data_pc(time, pc, PLO); 
            if (cw & PCH_IN) set_data_Z_pc(time, PHI); else set_data_pc(time, pc, PHI);
        }
        else {
            // bit set/reset on port c
            set_data_pc(time, datain & 1 ? 0xFF : 0, 1 << ((datain >> 1) & 7));
            if (logging & 4) instance->log("i8255: set PC bit: PC%u=%u", ((datain >> 1) & 7), datain & 1);
        }
    }
}

void I8255::pin_RD_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_reset->isinactive()) return;
    if (!pin_cs->isinactive()) return;
    if (pin_rd->isposedge()) {
        set_data_Z(time + 25000, DB);
        return;
    }
    if (pin_rd->isnegedge()) {
        BYTE addr = 0;
        if (pin_a0->isactive()) addr |= 1;
        if (pin_a1->isactive()) addr |= 2;
        BYTE dataout;
        if (addr == 0) {
            pa = get_data(PA);
            if (logging & 1) instance->log("i8255: read %s: 0x%02X", "PA", pa);
            dataout = pa;
        }
        else if (addr == 1) {
            pb = get_data(PB);
            if (logging & 2) instance->log("i8255: read %s: 0x%02X", "PB", pb);
            dataout = pb;
        }
        else if (addr == 2) {
            pc = get_data(PC);
            if (logging & 4) instance->log("i8255: read %s: 0x%02X", "PC", pc);
            dataout = pc;
        }
        else {
            if (logging) instance->log("i8255: read %s: 0x%02X", "CW", cw);
            dataout = cw;
        }
        set_data(time, dataout, DB);
    }
}


void I8255::simulate(ABSTIME time, DSIMMODES mode) {   
 
}   
       
      
void I8255::callback(ABSTIME time, EVENTID eventid)   {   

}   
              
