#define WIN32_LEAN_AND_MEAN             /* Exclude rarely-used stuff from Windows headers */

#include <intrin.h>
#include <windows.h>
#include <shlwapi.h>
#include <tchar.h>
#include "vsm.hpp"
#include "header.h"
#include "resource.h"



#define FAKE



////

#include "font.h"



#define ADD_X 64U
#define ADD_F 0

extern BYTE* ram;


////
#define VE_fl 0x04 /*video enable flag*/
#define IE_fl 0x40 /*interrupt enable flag*/
#define IR_fl 0x20 /*interrupt request flag*/

// no. of character clocks between DMA requests
const BYTE dma_sss[8] = { 0, 7, 15, 23, 31, 39, 47, 55 };



#pragma optimize("s", on)

BYTE I8275::model_id() {
    return 75;
}

I8275::~I8275() {
    if (ev) {
        SetEvent(ev);
        if (th) {
            if (hw) PostMessage(hw, WM_USER + 2, 0, 0);
            DWORD w = WaitForSingleObject(th, 5555);
            if (w == WAIT_TIMEOUT) {
                MessageBoxA(GetForegroundWindow(), "Thread waiting timeout.", "TV DISPLAY ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
                TerminateThread(th, 0);
            }
            CloseHandle(th);
        }
        CloseHandle(ev);
    }
    aligned_zfree(pbuff);
    zfree(bmi);
}


BOOL I8275::indicate(REALTIME time, ACTIVEDATA* data) {
    return false;
}


int I8275::isdigital(CHAR* pinname) {
    return 1;
}


DWORD WINAPI ThreadStart(I8275* p) {
    return p->Thread();
}

//(1000000000000 = E8 D4A51000) / clk
__declspec(naked) RELTIME __fastcall clk2ps(LONG) {
    __asm {
        xor     edx, edx
        mov     eax, 0E8h
        div     ecx
        push    eax
        mov     eax, 0D4A51000h
        div     ecx
        pop     edx
        ret
    }
}

/********************************************************************/
void I8275::setup(IINSTANCE* inst, IDSIMCKT* dsim) {
    instance = inst;
    //instance->log("START");
    //ckt = dsim;

    logg = BYTE(inst->getinitval("LOGGING", 1));
    LONG clk = instance->getinitval("CLK", 0);
    iclk = false;
    if (clk) {
        iclk = true;
        dsim->setclockcallback(0, clk2ps(clk), this, (CALLBACKHANDLERFN)&I8275::callback, 0);
        instance->log("i8275: internal clk set to %u Hz.", clk);
    }

    __stosd((unsigned long*)row_buf, 0, sizeof(row_buf) / 4);



    pin_CLK = instance->getdsimpin("CCLK", true);
    pin_cs = instance->getdsimpin("$CS$", true);
    pin_nRD = instance->getdsimpin("$RD$", true);
    pin_nWR = instance->getdsimpin("$WR$", true);
    pin_A0 = instance->getdsimpin("A0", true);

    pin_cs->sethandler(this, (PINHANDLERFN)&I8275::pin_CS_onchange);
    pin_nRD->sethandler(this, (PINHANDLERFN)&I8275::pin_RD_onchange);
    pin_nWR->sethandler(this, (PINHANDLERFN)&I8275::pin_WR_onchange);
    pin_CLK->sethandler(this, (PINHANDLERFN)&I8275::pin_CLK_onchange);

    char d[4] = "D0";
    for (unsigned i = 0; i < 8; i++) {
        pin_D[i] = instance->getdsimpin(d, true);
        d[1]++;
    }
    pin_IRQ = instance->getdsimpin("IRQ", true);

#ifndef FAKE
    pin_DRQ = instance->getdsimpin("DRQ", true);
    pin_nDACK = instance->getdsimpin("$DACK$", true);
    pin_LPEN = instance->getdsimpin("LPEN", true);
    pin_LTEN = instance->getdsimpin("LTEN", true);
    pin_LC0 = instance->getdsimpin("LC0", true);
    pin_LC1 = instance->getdsimpin("LC1", true);
    pin_LC2 = instance->getdsimpin("LC2", true);
    pin_LC3 = instance->getdsimpin("LC3", true);
    pin_CC0 = instance->getdsimpin("CC0", true);
    pin_CC1 = instance->getdsimpin("CC1", true);
    pin_CC2 = instance->getdsimpin("CC2", true);
    pin_CC3 = instance->getdsimpin("CC3", true);
    pin_CC4 = instance->getdsimpin("CC4", true);
    pin_CC5 = instance->getdsimpin("CC5", true);
    pin_CC6 = instance->getdsimpin("CC6", true);
    pin_LA0 = instance->getdsimpin("LA0", true);
    pin_LA1 = instance->getdsimpin("LA1", true);
    pin_VSP = instance->getdsimpin("VSP", true);
    pin_HLGT = instance->getdsimpin("HLGT", true);
    pin_RVV = instance->getdsimpin("RVV", true);
    pin_GPA0 = instance->getdsimpin("GPA0", true);
    pin_GPA1 = instance->getdsimpin("GPA1", true);
    pin_HRTC = instance->getdsimpin("HRTC", true);
    pin_VRTC = instance->getdsimpin("VRTC", true);
#endif
    pin_IRQ->drivestate(0, SLO);


#ifndef FAKE
    pin_DRQ->drivestate(0, SLO);
    pin_LTEN->drivestate(0, SLO);
    pin_LC0->drivestate(0, SLO);
    pin_LC1->drivestate(0, SLO);
    pin_LC2->drivestate(0, SLO);
    pin_LC3->drivestate(0, SLO);
    pin_CC0->drivestate(0, SLO);
    pin_CC1->drivestate(0, SLO);
    pin_CC2->drivestate(0, SLO);
    pin_CC3->drivestate(0, SLO);
    pin_CC4->drivestate(0, SLO);
    pin_CC5->drivestate(0, SLO);
    pin_CC6->drivestate(0, SLO);
    pin_LA0->drivestate(0, SLO);
    pin_LA1->drivestate(0, SLO);
    pin_VSP->drivestate(0, SLO);
    pin_HLGT->drivestate(0, SLO);
    pin_RVV->drivestate(0, SLO);
    pin_GPA0->drivestate(0, SLO);
    pin_GPA1->drivestate(0, SLO);
    pin_HRTC->drivestate(0, SHI);
    pin_VRTC->drivestate(0, SHI);
#endif
    //////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////////

    CREATEPOPUPSTRUCT cp;
    cp.caption = "TV DISPLAY";
    cp.id = 0;
    cp.type = PWT_USER;
    cp.flags = PWF_VISIBLE;
    cp.height = 0;
    cp.width = 0;
    pup = (IUSERPOPUP*)instance->createpopup(&cp);
    if (pup) {
        pup->setmsghlr(this);
    }
    else {
        instance->warning("Unable to create popup window.");
        return;
    }

    pbuff = aligned_zalloc((size_x + 16) * (size_y + 1024) * 4);
    if (!pbuff) {
        instance->fatal("malloc failed.");
        return;
    }

    bmi = (BITMAPINFO*)zalloc(sizeof(BITMAPINFO) + 256 * sizeof(COLOUR));
    if (!bmi) {
        instance->fatal("malloc failed.");
        return;
    }
    bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi->bmiHeader.biWidth = (size_x * 2);
    bmi->bmiHeader.biPlanes = 1;
    bmi->bmiHeader.biBitCount = 8;
    COLOUR* cl = (COLOUR*)&bmi->bmiColors;
            //  rrggbb

    cl[1] = 0x0099BBFF;
    cl[2] = 0x00666779;
    cl[3] = 0x00FFFFFF;
    cl[4] = 0x00444558;

    cl[9] = 0x005566AA;
    cl[10] = 0x00333446;
    cl[11] = 0x00AAAAAA;
    cl[12] = 0x00222336;

    ev = CreateEvent(nullptr, false, false, nullptr);
    if (ev == 0) {
        instance->fatal("CreateEvent failed.");
        return;
    }
    th = CreateThread(nullptr, 0, (LPTHREAD_START_ROUTINE)ThreadStart, this, 0, nullptr);
    if (th == 0) {
        instance->fatal("CreateThread failed.");
        return;
    }
    if (WaitForSingleObject(ev, 5555) == WAIT_TIMEOUT) {
        instance->warning("Thread timeout.");
    }
    while (hw == 0) {
        if (WaitForSingleObject(th, 10) != WAIT_TIMEOUT) {
            DWORD e = 0;
            GetExitCodeThread(th, &e);
            instance->error("Thread error %u.", e);
            break;
        }
    }
}

#pragma optimize("t", on)
/********************************************************************/

//    Z 
void I8275::set_data_Z(ABSTIME time) {
    for (unsigned i = 0; i < 8; i++) pin_D[i]->drivestate(time, FLT);
}


//    
BYTE I8275::get_data(void) {
    BYTE tmp = 0;
    BYTE b = 1;
    for (unsigned i = 0; i < 8; i++, b *= 2) if (!islow(pin_D[i]->istate())) tmp |= b;
    return tmp;
}



//   
void I8275::set_data(ABSTIME time, BYTE tmp) {
    BYTE b = 1;
    for (unsigned i = 0; i < 8; i++, b *= 2) pin_D[i]->drivebool(time, tmp & b);
}


void I8275::pin_CS_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_cs->isposedge()) set_data_Z(time);
}

void I8275::pin_RD_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_cs->isinactive()) return;
    if (pin_A0->isinactive()) {
        if (takt == 5) {
            if (pin_nRD->isnegedge()) {
                set_data(time, lp_char_numb_reg);
                if (logg & 2) instance->log("i8275: read lp_char_numb_reg %u", lp_char_numb_reg);
            }
            if (pin_nRD->isposedge()) {
                set_data_Z(time);
                takt++;
            }
        }
        else if (takt == 6) {
            if (pin_nRD->isnegedge()) {
                set_data(time, lp_row_numb_reg);
                if (logg & 2) instance->log("i8275: read lp_row_numb_reg %u", lp_row_numb_reg);
            }
            if (pin_nRD->isposedge()) {
                set_data_Z(time);
                takt = 0;
            }
        }
        else {
            if (logg && pin_nRD->isnegedge()) instance->warning("i8275: invalid rd %u", takt);
            set_data_Z(time);
        }
        return;
    }

    // SREG,    ,   return
    if (pin_nRD->isnegedge()) {
        set_data(time, SREG_reg);
        if (logg & 2) instance->log("i8275: read SREG_reg: 0x%02X", SREG_reg);
        if (SREG_reg & IR_fl) {
            SREG_reg &= ~IR_fl;
            pin_IRQ->drivebool(time, 0);
        }
        return;
    }
    if (pin_nRD->isposedge()) {
        set_data_Z(time + 25000);
        return;
    }
}


void I8275::pin_WR_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_cs->isinactive()) {

        //   
        if (pin_A0->isactive() && pin_nWR->isposedge()) {
            BYTE tmp = get_data();

            //      /( )2
            if ((tmp & 0xE0) == 0x20) {
                //start display cmd, next 0 data param
                display_cmd = tmp;
                if (logg) instance->log("i8275: start display cmd: 0x%02X", tmp);
                SREG_reg |= VE_fl;
                SREG_reg |= IE_fl;
                takt = 0;
            }
            else {
                switch (tmp)
                {
                case 0: //reset cmd, next 4 data param
                    takt = 1;//2,3,4   //   
                    SREG_reg &= ~VE_fl;
                    SREG_reg &= ~IE_fl;
                    pin_IRQ->drivebool(time, 0);
                    display_cmd = 0;

                    break;
                case 0x40: //stop display cmd, next 0 data param
                    takt = 0;
                    SREG_reg &= ~VE_fl;
                    SREG_reg &= ~IE_fl;
                    if (logg) instance->log("i8275: stop display cmd");

                    break;
                case 0x60: //read light pen cmd, next 2 data param
                    takt = 5;//6   //   
                    break;
                case 0x80: //load cursor pos cmd, next 2 data param
                    takt = 7;//8   //   
                    break;
                case 0xA0: //enable int cmd, next 0 data param
                    SREG_reg |= IE_fl;
                    takt = 0;
                    if (logg) instance->log("i8275: enable int cmd");
                    break;
                case 0xC0: //disable int cmd, next 0 data param
                    if (logg) instance->log("i8275: disable int cmd");
                    SREG_reg &= ~IE_fl;
                    takt = 0;
                    break;
                case 0xE0: //preset counters cmd, next 0 data param
                    takt = 0;
                    if (logg) instance->log("i8275: preset counters cmd");
                    break;
                default:
                    instance->warning("i8275: invalid command: 0x%02X", tmp);
                }
            }
            return;
        }


        if (pin_A0->isinactive() && pin_nWR->isposedge()) {

            //      ( )
            BYTE d = get_data();
            switch (takt) {
            case 1:
                screen_comp1_reg = d;
                if (logg) instance->log("i8275: reset cmd, screen_comp1: S=%u H=%u", d >> 7, d & 0x7F);
                takt++;
                break;
            case 2:
                screen_comp2_reg = d;
                if (logg) instance->log("i8275: reset cmd, screen_comp2: V=%u R=%u", d >> 6, d & 0x3F);
                takt++;
                break;
            case 3:
                screen_comp3_reg = d;
                if (logg) instance->log("i8275: reset cmd, screen_comp3: U=%u L=%u", d >> 4, d & 15);
                takt++;
                break;
            case 4:
                screen_comp4_reg = d;
                if (logg) {
                    instance->log("i8275: reset cmd, screen_comp4: M=%u F=%u C=%u Z=%u", d >> 7, (d >> 6) & 1, (d >> 4) & 3, d & 15);
                }
                takt = 0;
                _InterlockedOr8(&redraw, 2);
                break;
            case 7:
                cur_char_numb_reg = d;
                if (logg & 2) instance->log("i8275: load cursor pos cmd, char: %u", d);
                takt++;
                break;
            case 8:
                cur_row_numb_reg = d;
                if (logg & 2) instance->log("i8275: load cursor pos cmd, row: %u", d);
                takt = 0;
                break;
            default:
                instance->warning("i8275: invalid wr %u", takt);
            }
        } //
        return;
    }
#ifndef FAKE
    //  ,      DMA     WR _/-
    //                                                                         DMA
    if (pin_nDACK->isinactive() && (display_cmd & 0x20) && pin_nWR->isposedge()) {

        if (logg & 8) instance->log("i8275: dma_burst_cnt %u   row_buf_cnt[%u] %u", dma_burst_cnt, current_row, row_buf_cnt[current_row]);

        row_buf[row_buf_cnt[current_row]][current_row] = get_data();
        row_buf_cnt[current_row]++;
        dma_burst_cnt++;

        //     DMA 
        BYTE tmp = 1 << (display_cmd & 3);

        //,   DMA  ,     
        if ((dma_burst_cnt == tmp) || (row_buf_cnt[current_row] == ((screen_comp1_reg & 0x7f) + 1))) {

            //   DMA  ,    DMA
            pin_DRQ->drivestate(time, SLO);

            //          
            if (row_buf_cnt[current_row] != ((screen_comp1_reg & 0x7f) + 1)) {
                clk_between_dma_cnt = 1; // 
            }
            else {
                clk_between_dma_cnt = 0;// ,    
                if (logg & 8) instance->log("i8275: row_buf_cnt[%u] %u    dma_burst_cnt %u", current_row, row_buf_cnt[current_row], dma_burst_cnt);
            }
            dma_burst_cnt = 0;
        }
    }
#endif
}

//    ,     mode0/mode1
void I8275::set_LC04(ABSTIME time, BYTE tmp) {
    if (screen_comp4_reg & 0x80) {
        // mode1
        if (tmp == 0) {
            tmp = screen_comp3_reg & 0x0F;
        }
        else {
            if (tmp == (screen_comp3_reg & 0x0F)) tmp = 0;
            else tmp--;
        }
    }
    pin_LC0->drivebool(time, tmp & 1);
    pin_LC1->drivebool(time, tmp & 2);
    pin_LC2->drivebool(time, tmp & 4);
    pin_LC3->drivebool(time, tmp & 8);
}



//   
void I8275::set_CC06(ABSTIME time, BYTE tmp) {
    if (0x80 & tmp) {
        if (logg & 16) instance->log("i8275: attribute < : 0x%02X", tmp);
        //  pin_VSP->drivestate(time,1,SHI);
        //  Atrib_blank=1;
        pin_HLGT->drivebool(time, tmp & 1);
        pin_GPA0->drivebool(time, tmp & 4);
        pin_GPA1->drivebool(time, tmp & 8);
        pin_RVV->drivebool(time, tmp & 16);
        //tmp = 0;
    }
    else {
        //  if (Atrib_blank) {
        if (logg & 16) instance->log("i8275: attribute > : 0x%02X", tmp);
        //	  pin_VSP->drivestate(time,1,SLO);Atrib_blank=0;
          //}
        pin_CC0->drivebool(time, tmp & 1);
        pin_CC1->drivebool(time, tmp & 2);
        pin_CC2->drivebool(time, tmp & 4);
        pin_CC3->drivebool(time, tmp & 8);
        pin_CC4->drivebool(time, tmp & 16);
        pin_CC5->drivebool(time, tmp & 32);
        pin_CC6->drivebool(time, tmp & 64);
    }
}


void I8275::pin_CLK_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_CLK->isnegedge()) {
        if (iclk == 0) callback(time, 0);
    }
}


void I8275::callback(ABSTIME time, EVENTID eventid) {
    //       CLK       
    if (display_cmd & 0x20) {

        //   
        if (hrtc_cnt == 0) {
#ifndef FAKE
            pin_HRTC->drivestate(time, SLO);

            //          ( )
            if (row_cnt < ((screen_comp2_reg & 0x3F) + 1)) pin_VSP->drivestate(time, SLO);
#endif

            //  
            if (row_cnt == 0) {
#ifndef FAKE
                pin_VRTC->drivestate(time, SLO);
                set_LC04(time, LC04_cnt);
#endif
                char_cnt = 0;
            }
#ifndef FAKE
            //          7
            //(   ,    >6),        
            if ((row_cnt < ((screen_comp2_reg & 0x3F) + 1)) && (((screen_comp3_reg >> 4) & 0x0F) > 7)) {

                //     
                if ((LC04_cnt == 0) || (LC04_cnt == (screen_comp3_reg & 0x0f)))
                    pin_VSP->drivestate(time, SHI);
                else
                    pin_VSP->drivestate(time, SLO);
            }
#endif
        }

        //        
        if (hrtc_cnt < ((screen_comp1_reg & 0x7F) + 1)) {
#ifndef FAKE
            //   ,    ,  
            if (current_row) set_CC06(time, row_buf[char_cnt][0]);
            else set_CC06(time, row_buf[char_cnt][1]);

            //    ,    ,      
            //0x10 - 1/16  ,       
            if ((LC04_cnt == (screen_comp3_reg & 0x0F)) && (row_cnt == cur_row_numb_reg) && (cursor_blink_cnt & 0x02)) {
                if (char_cnt == cur_char_numb_reg) {
                    if (logg & 4) instance->log("i8275: char_cnt %u   row_cnt   %u   cursor_blink_cnt %u", char_cnt, row_cnt, cursor_blink_cnt);

                    //    ,  , , ,    
                    pin_VSP->drivestate(time, SLO);

                    // 
                    pin_LTEN->drivestate(0, SHI);
                }
                if (char_cnt == (cur_char_numb_reg + 1)) {
                    if (logg & 4) instance->log("i8275: char_cnt %u   row_cnt   %u", char_cnt, row_cnt);

                    //    ,         7
                    if (((screen_comp3_reg >> 4) & 0x0F) > 7) pin_VSP->drivestate(time, SHI);
                    pin_LTEN->drivestate(0, SLO);
                }
            }
            if (logg & 4) instance->log("i8275: char_cnt %u", char_cnt);
#endif
            //    
            char_cnt++;
        }

        //     
        hrtc_cnt++;

        //     
        if (hrtc_cnt == ((screen_comp1_reg & 0x7F) + 1)) {

#ifndef FAKE
            pin_HRTC->drivestate(time, SHI);

            //     
            pin_VSP->drivestate(time, SHI);
#endif
            LC04_cnt++;

            //        ...
            if (LC04_cnt == ((screen_comp3_reg & 0x0F) + 1)) {
                LC04_cnt = 0;
                row_cnt++;

                //    
                if (row_cnt == ((screen_comp2_reg & 0x3F) + 1)) {
#ifndef FAKE
                    pin_VRTC->drivestate(time, SHI);
                    pin_VSP->drivestate(time, SHI);
#endif
                    cursor_blink_cnt++;
                }

                //    ,  +    
                if (row_cnt == (((screen_comp2_reg & 0x3F) + 1) + (((screen_comp2_reg >> 6) & 0x03) + 1))) {
                    row_cnt = 0;
                    _InterlockedOr8(&redraw, 1);
                    SetEvent(ev);
                }

                //           
                if (row_cnt == (screen_comp2_reg & 0x3F)) {
                    if (SREG_reg & IE_fl) {
                        SREG_reg |= IR_fl;
                        pin_IRQ->drivebool(time, 1);

                    }
                }
#ifndef FAKE
                //          DMA  
                if (row_cnt < (screen_comp2_reg & 0x3F)) {

                    //   DMA   (     ),     DMA
                    if (row_buf_cnt[current_row]) {
                        if (SREG_reg & VE_fl) {
                            pin_DRQ->drivestate(time, SHI);
                        }
                        //    
                        if (current_row) current_row = 0; else current_row = 1;

                        //,         ,   DMA
                        row_buf_cnt[current_row] = 0;
                        if (logg & 4) instance->log("i8275: new row ---------- DMA reqw - row_cnt %u    current_row %u", row_cnt, current_row);
                    }
                }
                if (row_cnt == (((screen_comp2_reg & 0x3f) + 1) + ((screen_comp2_reg >> 6) & 0x03))) {
                    if (SREG_reg & VE_fl) {
                        pin_DRQ->drivestate(time, SHI);
                    }
                    if (current_row) current_row = 0; else current_row = 1;
                    row_buf_cnt[current_row] = 0;
                    if (logg & 4) instance->log("i8275: DMA reqw - new row  %u", current_row);
                }
#endif
            }// end if (LC04_cnt==((screen_comp3_reg&0x0f)+1))

#ifndef FAKE
            //   ,     mode0/mode1
            set_LC04(time, LC04_cnt);
#endif
            //     
            char_cnt = 0;
        }

        //      (   CLK)
        if (hrtc_cnt == (((screen_comp1_reg & 0x7F) + 1) + (((screen_comp4_reg & 0x0F) + 1) * 2))) hrtc_cnt = 0;
#ifndef FAKE
        //     DMA
        if (clk_between_dma_cnt) {
            clk_between_dma_cnt++;

            //  ,     DMA
            if (clk_between_dma_cnt == dma_sss[((display_cmd >> 2) & 7)] + 1) {
                clk_between_dma_cnt = 0;
                pin_DRQ->drivestate(time, SHI);
            }
        }
#endif
    }
}



//   
void I8275::simulate(ABSTIME time, DSIMMODES mode) {

}

void I8275::actuate(REALTIME time, ACTIVESTATE newstate) {

}


void I8275::runctrl(RUNMODES mode) {
    if (mode == RM_SUSPEND) {
        _InterlockedOr8(&redraw, 1);
        if (ev) SetEvent(ev);
    }
}


inline LRESULT I8275::WndProg(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam) {
    switch (msg) {
    case WM_PAINT:
        if (BeginPaint(hwnd, &ps)) {
            long b = ps.rcPaint.bottom;
            bmi->bmiHeader.biHeight = -b;
            long t = ps.rcPaint.top;
            long l = ps.rcPaint.left;
            SetDIBitsToDevice(ps.hdc, l, t, ps.rcPaint.right - l, b - t, l, t, t, b, pbuff, bmi, DIB_RGB_COLORS);
            EndPaint(hwnd, &ps);
        }
        return 0;
    case WM_ERASEBKGND:
        return 1;
    case WM_QUERYOPEN:
        redraw = 7;
        return 1;
    case WM_WINDOWPOSCHANGING:
    case WM_WINDOWPOSCHANGED:
        return 0;
    case WM_CLOSE:
        if (hwt) {
            PostMessage(hwt, WM_CLOSE, 0, 0);
            ShowWindow(hwnd, SW_HIDE);
        }
        return 0;
    case WM_DESTROY:
        hw = NULL;
        PostQuitMessage(0);
        break;
    case WM_USER + 1:
        redraw = 7;
        if (IsIconic(hwnd)) ShowWindow(hwnd, SW_RESTORE);
        else ShowWindow(hwnd, SW_SHOW);
        return 0;
    case WM_USER + 2:
        return DestroyWindow(hwnd);
    }
    return DefWindowProc(hwnd, msg, wParam, lParam);
}

LRESULT CALLBACK preWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    I8275* p = (I8275*)GetWindowLongPtr(hwnd, GWL_USERDATA);
    if (p) return p->WndProg(hwnd, msg, wParam, lParam);
    return DefWindowProc(hwnd, msg, wParam, lParam);
}


void I8275::draw() {
    BYTE* rp = ram;
    if (!rp) return;
    if (takt) return;
    char r = _InterlockedExchange8(&redraw, 0);
    if (!r) return;
    unsigned font_x = (screen_comp3_reg >> 4) + ADD_F;
    unsigned font_y = (screen_comp3_reg & 15) + ADD_F;
    if (r & 2) {
        unsigned x = font_x * ((screen_comp1_reg & 127) + 1U) + ADD_X;
        if (x & 1) x++;
        unsigned y = font_y * ((screen_comp2_reg & 63) + 1U);
        if ((x != size_x) || (y != size_y)) {
            if (x < 127) x = 127;
            if (y < 127) y = 127;
            size_x = x;
            size_y = y;
            bmi->bmiHeader.biWidth = (x * 2);
            pbuff = aligned_zrealloc(pbuff, (x + 16) * (y + 1024) * 4);
            int width = (size_x - ADD_X) * 2 + GetSystemMetrics(SM_CXFRAME) * 2 + 8;
            int height = size_y * 2 + GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION) + 17;
            SetWindowPos(hw, NULL, 0, 0, width, height, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOREDRAW | SWP_NOCOPYBITS);
        }
    }
    unsigned attr = 0x100;
    for (unsigned y = 0; y < (screen_comp2_reg & 63) + 1U; y++) {
        for (unsigned x = 0; x < (screen_comp1_reg & 127) + 1U; x++) {
            BYTE s = *rp++;
            while (s & 0x80) {
                if (attr) {
                    //if (s != 0x80) instance->log("attribute X:%02d, Y:%02d, A:0x%02X", x, y, s);
                    if (s == 0x80) attr = 0x100;
                    if (s & 0x10) attr = 0x300;
                    if (s & 0x02) attr |= 4;
                    if (s == 0xF0) attr = 0;
                }
                s = *rp++;
            }
            unsigned cl = attr;
            if ((x == cur_char_numb_reg) && (y == cur_row_numb_reg) && (cursor_blink_cnt & 8)) cl |= 2;
            unsigned fi = s * 16;
            WORD* wp = (WORD*)((BYTE*)pbuff + (4 + x * font_x * 2) + (4 + y * font_y * 2) * size_x * 2);
            for (unsigned j = 0; j < font_y; j++, fi++, wp += size_x * 2 - font_x) {
                BYTE f = font[fi];
                for (unsigned i = 0; i < font_x; i++, f = f >> 1, wp++) {
                    unsigned p = ((cl >> (8 * (f & 1))) & 255) * 257U;
                    wp[0] = LOWORD(p);
                    wp[size_x] = LOWORD(p | 0x0808);
                }
            }
        }
    }
    InvalidateRect(hw, nullptr, false);
}




extern "C" void* __ImageBase;

DWORD I8275::Thread() {
    DWORD e = 0;
    WNDCLASS wc;
    wc.style = 0;
    wc.lpfnWndProc = WNDPROC(preWndProc);
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = HINSTANCE(&__ImageBase);
    wc.hIcon = HICON(LoadImage(wc.hInstance, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 0, 0, LR_SHARED));
    wc.hCursor = HCURSOR(LoadImage(nullptr, IDC_ARROW, IMAGE_CURSOR, 0, 0, LR_SHARED));
    wc.hbrBackground = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = (TCHAR*)zalloc(8 * sizeof(TCHAR));
    if (!wc.lpszClassName) {
        SetEvent(ev);
        return ERROR_OUTOFMEMORY;
    }
    int cn = int(this);
    for (unsigned i = 0; i < 6; i++, cn >>= 6) TCHAR(wc.lpszClassName[i]) = 48 + (cn & 63) + i;
    DWORD_PTR c = RegisterClass(&wc);
    if (c) {
        DWORD st = WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU | WS_THICKFRAME;
        hw = CreateWindowEx(WS_EX_TOPMOST, (TCHAR*)c, _T("TV DISPLAY"), st, CW_USEDEFAULT, 0, 127, 127, 0, 0, wc.hInstance, nullptr);
        if (hw) {
            SetWindowLongPtr(hw, GWL_USERDATA, LONG_PTR(this));
            SetEvent(ev);
            MSG msg = { 0 };
            for (;;) {
                DWORD w = MsgWaitForMultipleObjectsEx(1, &ev, INFINITE, QS_ALLINPUT, 0);
                if (w == WAIT_OBJECT_0 + 1) {
                    while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) DispatchMessage(&msg);
                    if (msg.message == WM_QUIT) break;
                }
                else if (w == WAIT_OBJECT_0) {
                    draw();
                }
                else {
                    break;
                }
            }
        }
        e = GetLastError();
        if (hw) DestroyWindow(hw);
        UnregisterClass((TCHAR*)c, wc.hInstance);
    }
    else {
        e = GetLastError();
    }
    if (e) {
        char str[32];
        wvnsprintfA(str, 32, "Error: %u.", va_list(&e));
        MessageBoxA(GetForegroundWindow(), str, "TV DISPLAY ERROR", MB_ICONERROR | MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    }
    zfree((void*)wc.lpszClassName);
    SetEvent(ev);
    return e;
}





LRESULT I8275::msghlr(HWND hwnd, MESSAGE msg, WPARAM wParam, LPARAM lParam) {
    if (msg == WM_SHOWWINDOW) {
        if (hw && wParam) {
            PostMessage(hw, WM_USER + 1, wParam, lParam);
            return 0;
        }
    }
    else if (msg == WM_WINDOWPOSCHANGING) {
        if (((PWINDOWPOS)lParam)->flags & SWP_SHOWWINDOW) {
            ((PWINDOWPOS)lParam)->flags &= ~SWP_SHOWWINDOW;
            ((PWINDOWPOS)lParam)->flags |= SWP_HIDEWINDOW | SWP_NOACTIVATE;
            return 0;
        }
    }
    else if (msg == WM_CREATE) {
        hwt = hwnd;
    }
    return pup->callwindowproc(msg, wParam, lParam);
}





