#define WIN32_LEAN_AND_MEAN

#include <intrin.h>
#include <windows.h>   
#include "vsm.hpp"   
#include "Header.h"   

#pragma intrinsic(memcpy)

BYTE* ram = nullptr;

#pragma optimize("s", on)

BYTE RAM::model_id() {
    return 16;
}

void RAM::setup(IINSTANCE* inst, IDSIMCKT* dsim) {
    instance = inst;
    //ckt = dsim;
    if (!nv) {
        if (ram) {
            instance->fatal("only one instance of VIDEORAM is allowed.");
            return;
        }
    }

    pin_ce = instance->getdsimpin("$CE$", true);
    pin_oe = instance->getdsimpin("$OE$", true);
    pin_we = instance->getdsimpin("$WE$", true);

    char pinname[4] = "D0\0";
    for (unsigned i = 0; i < 8; i++, pinname[1]++) pin_d[i] = instance->getdsimpin(pinname, true);

    memcpy(pinname, "A0\0", 4);
    for (unsigned i = 0; i < 10; i++, pinname[1]++) pin_a[i] = instance->getdsimpin(pinname, true);
    pinname[1] = '1';
    pinname[2] = '0';
    pin_a[10] = instance->getdsimpin(pinname, true);
    pin_ce->sethandler(this, (PINHANDLERFN)&RAM::pin_CE_onchange);
    pin_oe->sethandler(this, (PINHANDLERFN)&RAM::pin_OE_onchange);
    pin_we->sethandler(this, (PINHANDLERFN)&RAM::pin_WE_onchange);

    pram = (BYTE*)zalloc(2048);
    if (pram == nullptr) {
        instance->fatal("malloc failed.");
        return;
    }
    if (nv) {
        char* fname = instance->getstrval("FILE");
        if (fname) {
            file = CreateFileA(fname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, 0);
            if (file != INVALID_HANDLE_VALUE) {
                DWORD r = 0;
                if (ReadFile(file, pram, 2048, &r, nullptr)) {
                    instance->log("nvram: Reading %u bytes from %s", r, fname);

                }
                else {
                    instance->warning("nvram: File %s Read Error: %d.", fname, GetLastError());
                }
            }
            else {
                instance->warning("nvram: File %s Open Error: %d.", fname, GetLastError());
            }
        }
    }
    else {
        ram = pram;
    }
}


RAM::~RAM() {
    zfree(pram);
    if (nv) {
        if (file != INVALID_HANDLE_VALUE) CloseHandle(file);
    }
    else {
        ram = nullptr;
    }
}



int RAM::isdigital(CHAR* pinname) {
    return 1;
}

BOOL RAM::indicate(REALTIME time, ACTIVEDATA* data) {
    return false;
}

#pragma optimize("t", on)
void RAM::runctrl(RUNMODES mode) {
    if (mode != RM_ANIMATE) {
        if (nv & 2) {
            nv = 1;
            if (file != INVALID_HANDLE_VALUE) {
                SetFilePointer(file, 0, nullptr, FILE_BEGIN);
                DWORD w;
                WriteFile(file, pram, 2048, &w, nullptr);
            }
        }
    }
}

void RAM::actuate(REALTIME time, ACTIVESTATE newstate) {

}   
       
void RAM::set_data_Z(ABSTIME time) {
    for (unsigned i = 0; i < 8; i++) pin_d[i]->drivestate(time, FLT);
}

BYTE RAM::get_data(void) {
    BYTE tmp = 0;
    BYTE b = 1;
    for (unsigned i = 0; i < 8; i++, b *= 2) if (!islow(pin_d[i]->istate())) tmp |= b;
    return tmp;
}

unsigned RAM::get_addr(void) {
    unsigned tmp = 0;
    unsigned b = 1;
    for (unsigned i = 0; i < 11; i++, b *= 2) if (!islow(pin_a[i]->istate())) tmp |= b;
    return tmp;
}


void RAM::set_data(ABSTIME time, BYTE tmp) {
    BYTE b = 1;
    for (unsigned i = 0; i < 8; i++, b *= 2) pin_d[i]->drivebool(time, tmp & b);
}



void RAM::pin_CE_onchange(ABSTIME time, DSIMMODES mode) {
    if (pin_ce->isposedge()) {
        set_data_Z(time);
    }
}

void RAM::pin_OE_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_ce->isinactive()) return;
    if (pin_oe->isposedge()) {
        set_data_Z(time + 10000);
    }
    else if (pin_oe->isnegedge()) {
        unsigned a = get_addr();
        set_data(time, pram[a]);
    }
}

void RAM::pin_WE_onchange(ABSTIME time, DSIMMODES mode) {
    if (!pin_ce->isinactive()) return;
    if (pin_we->isposedge()) {
        unsigned a = get_addr();
        BYTE d = get_data();
        if (pram[a] != d) {
            pram[a] = d;
            if (nv & 1) nv = 3;
        }
    }
}


void RAM::simulate(ABSTIME time, DSIMMODES mode) {      

}   
       
      
void RAM::callback(ABSTIME time, EVENTID eventid) {   

}   

               
